########################################
# The contents of this file are subject to the MLX PUBLIC LICENSE version
# 1.0 (the "License"); you may not use this file except in
# compliance with the License.
# 
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
# the License for the specific language governing rights and limitations
# under the License.
# 
# The Original Source Code is "compClust", released 2003 September 03.
# 
# The Original Source Code was developed by the California Institute of
# Technology (Caltech).  Portions created by Caltech are Copyright (C)
# 2002-2003 California Institute of Technology. All Rights Reserved.
########################################

__all__ = ["Labeling", "GlobalLabeling", "GlobalWrapper", "ElementLabeling",
           "GlobalElementLabeling", "GlobalElementWrapper",
           "subsetByLabeling", "ClusteredLabeling"]

import types

from Labeling import Labeling, ClusteredLabeling
from GlobalLabeling import GlobalLabeling
from GlobalLabeling import GlobalWrapper
from ElementLabeling import ElementLabeling
from GlobalElementLabeling import GlobalElementLabeling
from GlobalElementLabeling import GlobalElementWrapper
from compClust.util.listOps import unique, unravel
from compClust.mlx.views import SubsetView


#############################################################################
#
# Helper functions which perform sophisticated operations leveraging
# labelings and dataset objects.
#
#############################################################################

def subsetByLabeling(ds, labeling, obj, name=None):
    """
    Creates and returns an arbitrary view based on a valid labeling of the
    dataset.

    If the labeling passed is not directly tied to the dataset object,
    the subset returned will be the subset of self which corresponds
    to the datapoints labeling in the dataset to which the labeling is
    attatched.

    The application of this behavior can be expressed in the following
    situation:

        Given a Dataset, perform two clusterings which result in two
        labelings.  Using these labelings, create two views, one per
        clustering of a particular cluster. Then create a labeling for the
        view which labels all the data.  Now apply this labeling to the
        second view to create a subset of common datapoints. The code to do
        this may be like the following.

        view1 = dataset.subset(labeling1, "cluster 2")
        view2 = dataset.subset(labeling2, "cluster 6")

        newlabeling1 = view1.createLabeling()
        newlabeling1.addLabelToRows("foo", range(view1.getNumRows()))

        commonGenes = view2.subset(newlabeling1, "foo")

    At the conclusion of this code, the commonGenes view will hold the
    data which is in view2 and also labeled "foo" in newlabeling1
    """
    
    #
    # Override the string->filename conversion
    #
    
    if type(obj) == types.StringType:
      obj = [obj]

    labels  = labeling._castLabels(obj)
    keylist = []

    #
    # flatten the label list
    #

    labels = unravel(labels)
    
    #
    # There are two cases -- it is direcly attatched to self, or is it
    # attatched to another dataset/view
    #
    # Fetch a local, context labeling, if None is returned, do the more
    # complicated case
    #
    
    labelings = ds.getLabelings()
    local     = labeling._getLocalRef(ds)
    
    if local is not None:
      
      #
      # Standard subset operation, the labeling is attatched to the dataset
      # in question
      #
        
      for label in labels:
        keylist += labeling.getKeysByLabel(label)
      
    else:

      #
      # Trickier case.  We want to find the common root dataset of this
      # dataset and the dataset associated with the labeling.  To do this
      # we will trace the UIDs to their common root datasets and then derive
      # the keys to subset on
      #
      # This is effectively a join on UID
      #

      for label in labels:
        keylist += labeling._findCommonKeys(ds, label)

    #
    # Remove duplicate keys, and put them in order
    #

    keys = unique(keylist)
    keys.sort()

    #
    # In the special case of keys == [], return an empty Dataset
    # and munge the values because Numeric.array([[]]) creates a matrix
    # of size 1 rows by 0 columns, not 0 by 0
    #
    # To reduce memory consumption, keep the first row & column to produce
    # a 1x1 dataset
    #

    if len(keys) == 0:
      keys = [0, ds.getNumRows]
      v = SubsetView(ds, keys, name=name)
      v.numRows = 0
      v.numCols = 0
    else:
      v = SubsetView(ds, keys, name=name)
      
    return v

def castToGlobalLabeling(lab, name=None, removeLocal=1):
  """cast the given labeling to a global label, if name is none the name of the local 
     labeling is percolated to the global labeling.  if removeLocal is set, the local 
     labeling is removed.  A globalWrapper analogous to the lab is returned. """
  ds = lab.getDataset()
  if name is None:
    name = lab.getName()
  l = GlobalWrapper(ds,name)
  l.labelFrom(lab)
  if removeLocal:
    ds.removeLabeling(lab)
  return(l)

